// UsbBt2.cpp 2012/12/18
#include "mbed.h"
#include "rtos.h"
#include "BaseUsbHost.h"
#define DEBUG
#include "BaseUsbHostDebug.h"
#define TEST
#include "BaseUsbHostTest.h"
#include "UsbBt2.h"

static int LE16(const uint8_t* d)
{
    return d[0] | (d[1] << 8);
}

bthci::bthci(ControlEp* ctlEp):m_pEpIntIn(NULL),m_pEpBulkIn(NULL),m_pEpBulkOut(NULL)
{
    if (ctlEp == NULL) { // root hub
        DBG_OHCI(LPC_USB->HcRhPortStatus1);
        TEST_ASSERT_FALSE(LPC_USB->HcRhPortStatus1 & 0x200);
        ctlEp = new ControlEp();
        TEST_ASSERT_TRUE(ctlEp);
    }
    m_ctlEp = ctlEp;
    bool r = check(ctlEp);
    TEST_ASSERT(r);
    ParseConfiguration(ctlEp);
    int rc = ctlEp->SetConfiguration(1);
    TEST_ASSERT_EQUAL(rc, USB_OK);
}

bool bthci::check(ControlEp* ctlEp)
{
    if (ctlEp == NULL) {
        return false;
    }
    uint8_t buf[18];
    int r = ctlEp->GetDescriptor(1, 0, buf, 8);
    if (r != USB_OK) {
        return false;
    }
    DBG_HEX(buf, 8);
    r = ctlEp->GetDescriptor(1, 0, buf, 18);
    if (r != USB_OK) {
        return false;
    }
    DBG_HEX(buf, 18);
    uint16_t vid = *reinterpret_cast<uint16_t*>(buf+8);
    uint16_t pid = *reinterpret_cast<uint16_t*>(buf+10);
    DBG("VID PID: %04X %04X\n", vid, pid);
    if (buf[4] == 0xe0 && buf[5] == 0x01 && buf[6] == 0x01) {
        return true;
    }
    return false;
}

int bthci::cmdSend(uint16_t op)
{
   return cmdSend(op, NULL, 0);
}

int bthci::cmdSend(uint16_t op, const uint8_t* data, int size)
{
    uint8_t* buf = new uint8_t[size+3];
    TEST_ASSERT(buf);
    if (buf == NULL) {
        return USB_ERROR;
    }
    buf[0] = op;
    buf[1] = op>>8;
    buf[2] = size;
    if (data) {
        memcpy(buf+3, data, size);
    }
    int rc = m_ctlEp->controlSend(USB_REQUEST_TYPE_CLASS, 0, 0, 0, buf, size+3);
    TEST_ASSERT(rc == USB_OK);
    delete[] buf;
    return rc;
}

int bthci::cmdSend(uint16_t op, const char* fmt, ...) 
{   
    va_list vl;
    va_start(vl, fmt);
    uint8_t* buf = new uint8_t[255];
    TEST_ASSERT(buf);
    if (buf == NULL) {
        return USB_ERROR;
    }
    buf[0] = op;
    buf[1] = op>>8;
    int pos = 3;
    char* name;
    int name_len;
    uint16_t h;
    BD_ADDR* bdaddr;
    for(int i = 0; fmt[i]; i++) {
        switch(fmt[i]) {
            case 's':
                name = va_arg(vl, char*);
                name_len = strlen(name)+1;
                memcpy(buf+pos, name, name_len);
                pos += name_len;
                break;
            case 'B':
                buf[pos++] = va_arg(vl, int);
                break;
            case 'H':
                h = va_arg(vl, int);
                buf[pos++] = h;
                buf[pos++] = h>>8;
                break;
            case 'A':
                bdaddr = va_arg(vl, BD_ADDR*);
                memcpy(buf+pos, bdaddr, 6);
                pos += 6;
                break;
            default:
                DBG("op=%04X fmt=%s i=%d", op, fmt, i);
                TEST_ASSERT(0);
                //break;
        }
    }
    buf[2] = pos-3;
    int rc = m_ctlEp->controlSend(USB_REQUEST_TYPE_CLASS, 0, 0, 0, buf, pos);
    TEST_ASSERT(rc == USB_OK);
    delete[] buf;
    return rc;
}

int bthci::eventReceive(uint8_t* buf, int size, int millisec)
{
    //DBG("%p buf=%p size=%d millisec=%d\n", this, buf, size, millisec);
    TEST_ASSERT(m_pEpIntIn);
    return m_pEpIntIn->interruptReceive(buf, size, millisec);
}

int bthci::AclSend(uint8_t* buf, int size, int millisec)
{
    TEST_ASSERT(m_pEpBulkOut);
    return m_pEpBulkOut->bulkSend(buf, size, millisec);
}

int bthci::AclReceive(uint8_t* buf, int size, int millisec)
{
    TEST_ASSERT(m_pEpBulkIn);
    return m_pEpBulkIn->bulkReceive(buf, size, millisec);
}

int bthci::ParseConfiguration(ControlEp* ctlEp)
{
    int rc;
    uint8_t ConfigDesc[9];
    int index = 0;
    rc = ctlEp->GetDescriptor(USB_DESCRIPTOR_TYPE_CONFIGURATION, index, ConfigDesc, sizeof(ConfigDesc));
    TEST_ASSERT(rc == USB_OK);
    DBG_BYTES("ConfigDescriptor 9bytes", ConfigDesc, sizeof(ConfigDesc));
    TEST_ASSERT(ConfigDesc[0] == 9);
    TEST_ASSERT(ConfigDesc[1] == 0x02);
    int wTotalLength = *((uint16_t*)&ConfigDesc[2]);
    DBG("TotalLength: %d\n", wTotalLength);
    int bConfigValue = ConfigDesc[5];
    TEST_ASSERT(bConfigValue == 1);
    DBG("ConfigValue: %d\n", bConfigValue);
    VERBOSE("MaxPower: %d mA\n", ConfigDesc[8]*2);   

    uint8_t* buf = new uint8_t[wTotalLength];
    TEST_ASSERT(buf);
    rc = ctlEp->GetDescriptor(USB_DESCRIPTOR_TYPE_CONFIGURATION, index, buf, wTotalLength);
    TEST_ASSERT(rc == USB_OK);
    TEST_ASSERT(ConfigDesc[1] == 0x02);
    for (int pos = 0; pos < wTotalLength; pos += buf[pos]) {
        DBG_BYTES("CFG", buf+pos, buf[pos]);
        int type = buf[pos+1];
        if (USB_DESCRIPTOR_TYPE_INTERFACE == type) { // 0x04
            DBG("InterfaceNumber: %d\n", buf[pos+2]);
            DBG("AlternateSetting: %d\n", buf[pos+3]);
            DBG("NumEndpoint: %d\n", buf[pos+4]);
            VERBOSE("InterfaceClass: %02X\n", buf[pos+5]);
            VERBOSE("InterfaceSubClass: %02X\n", buf[pos+6]);
            VERBOSE("InterfaceProtocol: %02X\n", buf[pos+7]);
            TEST_ASSERT(buf[pos+6] == 0x01);
            TEST_ASSERT(buf[pos+7] == 0x01);
        } 
        if (USB_DESCRIPTOR_TYPE_ENDPOINT == type) {
            TEST_ASSERT(buf[pos] == 7);
            uint8_t att = buf[pos+3];
            uint8_t ep = buf[pos+2];
            uint16_t size = LE16(buf+pos+4);
            DBG("EndpointAddress: %02X\n", ep);
            DBG("Attribute: %02X\n", att);
            DBG("MaxPacketSize: %d\n", size);
            int addr = ctlEp->GetAddr(); 
            if (att == 3) { // interrupt
                if (m_pEpIntIn == NULL) {
                    m_pEpIntIn = new InterruptEp(addr, ep, size);
                }
            } else if (att == 2) { // bulk
                if (ep & 0x80) { // IN
                  if (m_pEpBulkIn == NULL) {
                      m_pEpBulkIn = new BulkEp(addr, ep, size);
                  }
              } else { // OUT
                  if (m_pEpBulkOut == NULL) {
                      m_pEpBulkOut = new BulkEp(addr, ep, size);
                  }
              } 
          }
      }
      if (m_pEpIntIn && m_pEpBulkIn && m_pEpBulkOut) { // cut off
          break;
      }
  }
  delete[] buf;
  TEST_ASSERT(m_pEpIntIn);
  TEST_ASSERT(m_pEpBulkIn);
  TEST_ASSERT(m_pEpBulkOut);
  return 0;   
}
